### Persefone.jl - a model of agricultural landscapes and ecosystems in Europe.
###
### This file is responsible for managing the crop growth modules.
###

#FIXME Looking at fields.pdf in the Persefone output, crop maturity is not recognised
# properly, so most crops (except maize) are never planted. At the same time, maize never
# seems to start growing.

#TODO write tests to compare our output to the output of the original ALMaSS algorithm

module ALMaSS

## The two input files required by ALMaSS. Must be present in the crop directory.
const CROPFILE = "crop_data_general.csv"
const GROWTHFILE = "crop_growth_curves.csv"

using Persefone:
    AnnualDate,
    Management,
    Length,
    cm,
    SimulationModel,
    fertiliser,
    maxtemp,
    mintemp

import Persefone:
    stepagent!,
    croptype,
    cropname,
    cropheight,
    cropcover,
    cropyield,
    sow!,
    harvest!,
    isharvestable,
    bounds

using Dates: Date, month, monthday
using CSV: CSV

"""
    GrowthPhase

ALMaSS crop growth curves are split into five phases, triggered by
seasonal dates or agricultural events.
"""
@enum GrowthPhase janfirst sow marchfirst harvest1 harvest2

"""
    CropCurveParams

The values in this struct define one crop growth curve.
"""
struct CropCurveParams
    #TODO add Unitful
    curveID::Int
    highnutrients::Bool
    GDD::Dict{GrowthPhase,Vector{Float64}}
    LAItotal::Dict{GrowthPhase,Vector{Float64}}
    LAIgreen::Dict{GrowthPhase,Vector{Float64}}
    height::Dict{GrowthPhase,Vector{Length{Float64}}}
end

"""
    CropType

The type struct for all crops. Currently follows the crop growth model as
implemented in ALMaSS.
"""
struct CropType
    #FIXME this needs thinking about. The sowing and harvest dates belong in the farm model,
    # not here. Also, we need to harmonise crops across the crop growth models.
    name::String
    group::String
    minsowdate::Union{Missing,AnnualDate}
    maxsowdate::Union{Missing,AnnualDate}
    minharvestdate::Union{Missing,AnnualDate}
    maxharvestdate::Union{Missing,AnnualDate}
    mingrowthtemp::Union{Missing,Float64}
    highnutrientgrowth::Union{Missing,CropCurveParams}
    lownutrientgrowth::Union{Missing,CropCurveParams}
end

cropname(ct::CropType) = ct.name

"""
    CropState

The state data for an ALMaSS vegetation point calculation.  Usually
part of a `FarmPlot`.
"""
mutable struct CropState
    #TODO add Unitful
    croptype::CropType
    phase::GrowthPhase
    growingdegreedays::Float64
    height::Length{Float64}
    LAItotal::Float64
    LAIgreen::Float64
    #biomass::Float64 #XXX I need to figure out how to calculate this
    mature::Bool #TODO how do we determine this?
    events::Vector{Management} #FIXME does this do anything?
end

croptype(cs::CropState) = cs.croptype
cropname(cs::CropState) = cropname(croptype(cs))
cropheight(cs::CropState) = cs.height
cropcover(cs::CropState) = 0.0  # TODO: related to LAItotal, LAIgreen?
cropyield(cs::CropState) = 0.0  # TODO: units? needs biomass?
isharvestable(cs::CropState) = cs.mature

"""
    Base.tryparse(type, str)

Extend `tryparse` to allow parsing GrowthPhase values.
(Needed to read in the CSV parameter file.)
"""
function Base.tryparse(::Type{GrowthPhase}, str::String)
    str == "janfirst" ? janfirst :
        str == "sow" ? sow :
        str == "marchfirst" ? marchfirst :
        str == "harvest1" ? harvest1 :
        str == "harvest2" ? harvest2 :
        nothing
end

"""
    buildgrowthcurve(data)

Convert a list of rows from the crop growth data into a CropCurveParams object.
"""
function buildgrowthcurve(data::Vector{CSV.Row})
    isempty(data) && return missing
    GDD = Dict(janfirst=>Vector{Float64}(), sow=>Vector{Float64}(),
               marchfirst=>Vector{Float64}(), harvest1=>Vector{Float64}(),
               harvest2=>Vector{Float64}())
    LAItotal = Dict(janfirst=>Vector{Float64}(), sow=>Vector{Float64}(),
                    marchfirst=>Vector{Float64}(), harvest1=>Vector{Float64}(),
                    harvest2=>Vector{Float64}())
    LAIgreen = Dict(janfirst=>Vector{Float64}(), sow=>Vector{Float64}(),
                    marchfirst=>Vector{Float64}(), harvest1=>Vector{Float64}(),
                    harvest2=>Vector{Float64}())
    height = Dict(janfirst=>Vector{Length{Float64}}(), sow=>Vector{Length{Float64}}(),
                  marchfirst=>Vector{Length{Float64}}(), harvest1=>Vector{Length{Float64}}(),
                  harvest2=>Vector{Length{Float64}}())
    for e in data        
        append!(GDD[e.growth_phase], e.GDD)
        append!(LAItotal[e.growth_phase], e.LAI_total)
        append!(LAIgreen[e.growth_phase], e.LAI_green)
        append!(height[e.growth_phase], e.height * cm)  # assumes `height` is in units of `cm`
    end
    CropCurveParams(data[1].curve_id, data[1].nutrient_status=="high",
                    GDD, LAItotal, LAIgreen, height)
end

"""
    readcropparameters(cropdirectory)

Parse a CSV file containing the required parameter values for each crop
(as produced from the original ALMaSS file by `convert_almass_data.py`).
"""
function readcropparameters(cropdirectory::String)
    @debug "Reading crop parameters"
    cropdata = CSV.File(joinpath(cropdirectory, CROPFILE), missingstring="NA",
                        types=[String,AnnualDate,AnnualDate,AnnualDate,AnnualDate,Float64,String])
    growthdata = CSV.File(joinpath(cropdirectory, GROWTHFILE), missingstring="NA",
                          types=[Int,String,String,GrowthPhase,String,
                                 Float64,Float64,Float64,Float64])
    croptypes = Dict{String,CropType}()
    for crop in cropdata
        cropgrowthdata = growthdata |> filter(x -> !ismissing(x.crop_name) &&
                                              x.crop_name == crop.name)
        highnuts = buildgrowthcurve(cropgrowthdata |>
                                    filter(x -> x.nutrient_status=="high"))
        lownuts = buildgrowthcurve(cropgrowthdata |>
                                   filter(x -> x.nutrient_status=="low"))
        # TODO: set crop group temporarily until there is a column in
        # the csv file
        crop_group = "CROP_GROUP_NOT_SET"
        croptypes[crop.name] = CropType(crop.name, crop_group, crop.minsowdate, crop.maxsowdate,
                                        crop.minharvestdate, crop.maxharvestdate,
                                        crop.mingrowthtemp, highnuts, lownuts)
    end
    croptypes
end

"""
    stepagent!(cropstate, model)

Update a farm plot by one day.
"""
function stepagent!(cs::CropState, model::SimulationModel)
    # update the phase on key dates
    if monthday(model.date) == (1, 1)
        cs.phase = ALMaSS.janfirst
        cs.growingdegreedays = 0.0
    elseif monthday(model.date) == (3, 1)
        cs.phase = ALMaSS.marchfirst
        cs.growingdegreedays = 0.0
    end
    # update growing degree days
    # if no crop-specific base temperature is given, default to 5°C
    # (https://www.eea.europa.eu/publications/europes-changing-climate-hazards-1/heat-and-cold/heat-and-cold-2014-mean)
    basetemp = cs.croptype.mingrowthtemp
    ismissing(basetemp) && (basetemp = 5.0)
    gdd = (maxtemp(model)+mintemp(model))/2 - basetemp
    gdd > 0 && (cs.growingdegreedays += gdd)
    # update crop growth
    growcrop!(cs, gdd, model)
end


## CROP MANAGEMENT AND GROWTH FUNCTIONS

"""
    sow!(cropstate, model, cropname)

Change the cropstate to sow the specified crop.
"""
function sow!(cs::CropState, model::SimulationModel, cropname::String)
    !ismissing(cs.croptype.minsowdate) && model.date < cs.croptype.minsowdate &&
        @warn "$(model.date) is earlier than the minimum sowing date for $(cropname)."
    cs.croptype = model.crops[cropname]
    cs.phase = ALMaSS.sow
    cs.growingdegreedays = 0.0
    cs.height = 0.0cm
    cs.LAItotal = 0.0
    cs.LAIgreen = 0.0
    cs.mature = false
    cs.events = Vector{Management}()
end

"""
    harvest!(cropstate, model)

Harvest the crop of this cropstate.
"""
function harvest!(cs::CropState, model::SimulationModel)
    cs.phase in [ALMaSS.harvest1, ALMaSS.harvest2] ?
        cs.phase = ALMaSS.harvest2 :
        cs.phase = ALMaSS.harvest1
    cs.growingdegreedays = 0.0
    cs.mature = false
    # height & LAI will be automatically adjusted by the growth function
    #TODO calculate and return yield
end

#TODO fertilise!()
#TODO spray!()
#TODO till!()

"""
    growcrop!(cropstate, gdd, model)

Apply the relevant crop growth model to update the plants crop state
on this farm plot.  Implements the ALMaSS crop growth model by Topping
et al. (see `ALMaSS/Landscape/plant.cpp:PlantGrowthData::FindDiff()` and
`ALMaSS/Landscape/elements.cpp:VegElement::DoDevelopment()`).
"""
function growcrop!(cs::CropState, gdd::Float64, model::SimulationModel)
    fertiliser in cs.events ?
        curve = cs.croptype.lownutrientgrowth :
        curve = cs.croptype.highnutrientgrowth
    points = curve.GDD[cs.phase] #FIXME what if the curve is empty?
    for p in 1:length(points)
        if points[p] == 99999
            !(cs.phase in (janfirst, sow)) && (cs.mature = true) #FIXME only in the last phase?
            return # the marker that there is no further growth this phase
        elseif points[p] == -1 # the marker to set all variables to specified values
            cs.height = curve.height[cs.phase][p]
            cs.LAItotal = curve.LAItotal[cs.phase][p]
            cs.LAIgreen = curve.LAIgreen[cs.phase][p]
            return
        else
            gdd = cs.growingdegreedays
            # figure out which is the correct slope value to use for growth
            if p == length(points) || gdd < points[p+1]
                #FIXME it appears to me from the ALMaSS source code that the curve value
                # for the day is multiplied with that day's growingdegreedays to give the
                # diff, which is then added to the previous values. However, this results
                # in values that are staggeringly too large. I'm not quite sure what the
                # issue is here...
                cs.height += bounds(curve.height[cs.phase][p]*gdd)
                cs.LAItotal += bounds(curve.LAItotal[cs.phase][p]*gdd)
                cs.LAIgreen += bounds(curve.LAIgreen[cs.phase][p]*gdd)
                return
            end
            #XXX To be precise, we ought to figure out if one or more inflection
            # points have been passed between yesterday and today, and calculate the
            # growth exactly up to the inflection point with the old slope, and onward
            # with the new slope. Not doing so will introduce a small amount of error,
            # although I think this is acceptable.
        end
    end
end

end # module ALMaSS
